#[macro_use]
extern crate slog;

use std::io;
use std::sync::{Arc, Mutex};
use std::thread;

use smol::Task;
use smol::net::TcpStream;

use tui::backend::{Backend, TermionBackend};
use tui::Terminal;
use termion::raw::IntoRawMode;
use termion::input::TermRead;
use termion::event::Key;

use futures::StreamExt;
use futures_signals::signal::SignalExt;

use clap::{App, Arg};

mod banner;
mod config;
mod app;
mod input;
mod util;
mod ui;
mod schema;

use banner::BANNER;

static DEFAULT_SERVER: &'static str = "localhost:59661";

fn main() -> Result<(), io::Error> {

    let matches = App::new("sute 🌸")
        .version(env!("CARGO_PKG_VERSION"))
        .author(env!("CARGO_PKG_AUTHORS"))
        .about(env!("CARGO_PKG_DESCRIPTION"))
        .usage("Press `?` in the GUI to see keybindings")
        .before_help(BANNER)
        .arg(Arg::with_name("config")
            .short("c")
            .long("config")
            .help("Specify configuration file path")
            .takes_value(true))
        .arg(Arg::with_name("server")
            .short("s")
            .long("server")
            .help("Connect to the specified address[:port] as server")
            .takes_value(true))
        .get_matches();

    // 
    let server = matches.value_of("server").unwrap_or(DEFAULT_SERVER);

    // Set up logging
    let drain = Arc::new(app::LogDrain::new());
    let log = slog::Logger::root(slog::Fuse::new(drain.clone()), o!());

    let lex = smol::LocalExecutor::new();

    let resize = util::Resize::new()?;
    let stream_f = async move {
        TcpStream::connect(server).await.unwrap()
    };

    let stream = smol::block_on(lex.run(stream_f));

    let (rpc_future, mut api) = schema::bootstrap(log.clone(), stream);

    let app = app::Sute::new(resize, log.clone(), drain, api);

    let mut stream = app.to_stream();

    let ui_future = async move {
        let stdout = io::stdout().into_raw_mode()?;
        let backend = TermionBackend::new(stdout);
        let mut terminal = Terminal::new(backend)?;
        terminal.hide_cursor()?;

        let mut tick_c = '|';

        // Refresh the screen once by resizing the terminal
        if let Ok((x,y)) = termion::terminal_size() {
            terminal.resize(tui::layout::Rect::new(0, 0, x,y)).unwrap();
        }

        loop {
            if let Some(mut state) = stream.next().await {
                if !state.running {
                    break;
                }

                tick_c = match tick_c {
                    '\\' => '|',
                    '|'  => '/',
                    '/'  => '-',
                    '-'  => '\\',
                    _ => '|',
                };
                state.tick_c = tick_c;
                terminal.draw(|f| ui::draw_ui(f, &mut state))?; 
            } else {
                break;
            }
        }

        // TODO: Ensure that will always be run
        terminal.show_cursor()?;

        Ok(())
    };

    //lex.spawn(rpc_future).detach();
    let r: Result<(), io::Error> = smol::block_on(lex.run(ui_future));
    //smol::block_on(lex.run(Box::pin(app)));

    Ok(())
}
